from sqlalchemy import DateTime, Float, Integer, String


def properties_from_coltype(sqlType):
    if isinstance(sqlType, Integer):
        return {
            'type': 'integer',
            'format': 'int64'
        }
    elif isinstance(sqlType, Float):
        return {
            'type': 'number'
        }
    elif isinstance(sqlType, String):
        return {
            'type': 'string'
        }
    elif isinstance(sqlType, DateTime):
        return {
            'type': 'string',
            'format': 'date-time'
        }
    return {}


def gen_response(http_method, resource_name):
    '''
    This function receives a HTTP method and a resource name and generates the standard responses of the method

    :param str http_method:
        http method used

    :param str resource_name:
        name of the resource being served

    :rtype dict:
    :return: dict containing the standard responses of the method
    '''
    if http_method == 'GET_Collection':
        return {
            '200': {
                'description': 'successful operation',
                'schema': {
                    'type': 'array',
                    'items': {
                        '$ref': '#/definitions/%s' % resource_name
                    }
                }
            }
        }
    elif http_method == 'PUT':
        return {
            '204': {
                'description': 'successfully updated'
            },
            '400': {
                'description': 'invalid %s supplied' % resource_name
            },
            '404': {
                'description': '%s not found' % resource_name
            }
        }
    elif http_method == 'GET':
        return {
            '200': {
                'description': 'successful operation',
                'schema': {
                    '$ref': '#/definitions/%s' % resource_name
                }
            },
            '400': {
                'description': 'invalid ID supplied'
            },
            '404': {
                'description': '%s not found' % resource_name
            }
        }
    elif http_method == 'POST':
        return {
            '201': {
                'description': 'the %s was created successfully' % resource_name
            },
            '405': {
                'description': 'invalid input'
            }
        }
    elif http_method == 'DELETE':
        return {
            '204': {
                'description': 'successfully deleted'
            },
            '400': {
                'description': 'invalid ID supplied'
            },
            '404': {
                'description': '%s not found' % resource_name
            }
        }

    return None


def gen_spec(model, http_method, is_child=False):
    '''
    This function receives a sqlalchemy class model and a HTTP method and generates the flasgger spec dict

    :param bool is_child:
        Tels if the model is a child resource in the route, meaning that a parent id must be part of the route

    :param class model:
        sqlachemy mapper class

    :param str http_method:
        http method used

    :rtype: dict
    :return: dict containing the spec
    '''
    res_name = model.__tablename__

    responses = gen_response(http_method, res_name)

    properties = {}
    for col in model.__table__.columns:
        if col.name != 'id':
            pro_name = col.name
            properties[pro_name] = properties_from_coltype(col.type)

    parameters = []
    if is_child:
        parameters.append({
            'name': 'parent_id',
            'in': 'path',
            'required': True,
            'type': 'integer',
            'format': 'int64'
        })
    if http_method == 'GET' or http_method == 'PUT' or http_method == 'DELETE':
        parameters.append({
            'name': 'id',
            'in': 'path',
            'required': True,
            'type': 'integer',
            'format': 'int64'
        })
    if http_method == 'POST' or http_method == 'PUT':
        parameters.append({
            'name': 'body',
            'in': 'body',
            'required': True,
            'schema': {
                '$ref': '#/definitions/%s' % res_name
            }
        })

    produces = ["application/json"]
    receives = ["application/json"]

    specs_dict = {
        "tags": [res_name],
        "parameters": parameters,
        "produces": produces,
        "receives": receives,
        "schemes": [
            "http",
            "https"
        ],
        "deprecated": False,
        "security": [
            {
                "Bearer": []
            }
        ],
        "responses": responses,
        "definitions": {
            res_name: {
                "type": "object",
                "properties": properties
            }
        }
    }

    return specs_dict



from sqlalchemy import desc, or_, and_
import json
import operator

# Filter operators defined on SqlAlchemy ColumnElement
SQLA_OPERATORS = {
    'like': 'like',
    'notlike': 'notlike',
    'ilike': 'ilike',
    'notilike': 'notilike',
    'is': 'is_',
    'isnot': 'isnot',
    'match': 'match',
    'startswith': 'startswith',
    'endswith': 'endswith',
    'contains': 'contains',
    'in': 'in_',
    'notin': 'notin_',
    'between': 'between',
    'eq': '__eq__',
    'ne': '__ne__',
    'gt': '__gt__',
    'ge': '__ge__',
    'lt': '__lt__',
    'le': '__le__',
}

def parse_value(value, serializer):
    if not serializer:
        return value
    if isinstance(value, list):
        return [serializer.load(item) for item in value]
    return serializer.load(value)


def get_operator(column, op_name, value, serializer):
    """
    :param column:
         SQLAlchemy ColumnElement

    :param op_name:
        Key of OPERATORS or COMPARE_OPERATORS

    :param value:
        value to be applied to the operator

    :rtype: ColumnOperators
    :return:
        returns a boolean, comparison, and other operators for ColumnElement expressions.
        ref: http://docs.sqlalchemy.org/en/latest/core/sqlelement.html#sqlalchemy.sql.operators.ColumnOperators
    """
    if not op_name:
        return column.operate(operator.eq, parse_value(value, serializer))
    elif op_name in SQLA_OPERATORS:
        op = SQLA_OPERATORS.get(op_name)
        if op == 'between':
            return column.between(parse_value(value[0], serializer), parse_value(value[1], serializer))
        return getattr(column, op)(parse_value(value, serializer))
    else:
        raise ValueError('Unknown operator {}'.format(op_name))


def get_field_serializer_or_none(serializer, field_name):
    field = serializer._fields.get(field_name)
    if not field:
        return None
    return field.serializer


def query_from_request(model, model_serializer, request, query=None):
    """
    Perform a filtered search in the database model table using query parameters in the http URL,
    disposed on the request args. The default logical operator is AND, but you can set the OR as
    in the following examples:

        a) OR -> ?filter={"$or":{"name": {"startswith": "Terrans 1"},"location": "Location 1"}}
        b) AND -> ?filter={"$and":{"name": {"ilike": "%Terrans 1%"},"location": "Location 1"}}
            or ?filter={"name": {"ilike": "%Terrans 1%"},"location": {"eq": "Location 1"}}

    It also paginate the response if a 'page' value is present in the query parameters. A 'per_page'
    value in the query parameters defines the page length, default to 20 items.

    Ordered search is available using 'order_by=<col_name>'. The minus sign ("-<col_name>") could be
    used to set descending order.

    :param class model:
        SQLAlchemy model class representing a database resource

    :param model_serializer:
        instance of model serializer

    :param request:
        Flask http request data

    :param query:
        SQLAlchemy query instance

    :rtype: list|dict
    :return: the serialized response: "if 'page' is defined in the query params, a dict with page, per page, count and results is returned,
    otherwise returns a list of serialized objects"
    """
    if not query:
        query = model.query

    def build_filter_operator(column_name, request_filter, serializer):
        if column_name == '$or':
            return or_(build_filter_operator(attr, value, serializer) for attr, value in request_filter.items())
        elif column_name == '$and':
            return and_(build_filter_operator(attr, value, serializer) for attr, value in request_filter.items())
        if isinstance(request_filter, dict):
            op_name = next(iter(request_filter))
            return get_operator(getattr(model, column_name), op_name, request_filter.get(op_name),
                                get_field_serializer_or_none(serializer, column_name))
        return get_operator(getattr(model, column_name), None, request_filter,
                            get_field_serializer_or_none(serializer, column_name))

    if 'filter' in request.args:
        filters = json.loads(request.args['filter'])
        for attr, value in filters.items():
            query = query.filter(build_filter_operator(attr, value, model_serializer))
    if 'limit' in request.args:
        limit = request.args['limit']
        query = query.limit(limit)
    if 'order_by' in request.args:
        fields = request.args['order_by'].split(',')
        for field in fields:
            if field[0] == '-':
                col = getattr(model, field[1:])
                query = query.order_by(desc(col))
            else:
                col = getattr(model, field)
                query = query.order_by(col)
    if 'page' in request.args:
        data = query.paginate()
        return {
            'page': data.page,
            'per_page': data.per_page,
            'count': data.total,
            'results': [model_serializer.dump(item) for item in data.items]
        }
    else:
        data = query.all()
        return [model_serializer.dump(item) for item in data]
